/**
 * @file TLC5947.c
 *
 * @date 27.04.2015
 * @author Robert Steigemann
 */ 

#include <util/delay.h>

#include "TLC5947.h"

// Macros for simple PIN Change
#define SET_XLAT (TLC5947_PORT|=(1<<TLC5947_XLAT))
#define CLR_XLAT (TLC5947_PORT&=(~(1<<TLC5947_XLAT)))
#define SET_BLANK (TLC5947_PORT|=(1<<TLC5947_BLANK))
#define CLR_BLANK (TLC5947_PORT&=(~(1<<TLC5947_BLANK)))

// Global Buffer for all LED levels, organized as 12-Bit Values
static uint8_t grayscale[TLC5947_CHIPS][36];

// Transfer a single byte via USART	
static inline void transmit_byte(uint8_t data)
{
#if (TLC5947_USART==0)	
	while(!(UCSR0A&(1<<UDRE0)));				// wait for empty transmit buffer
	UDR0 = data;								// send data
#elif (TLC5947_USART==1)
	while(!(UCSR1A&(1<<UDRE1)));
	UDR1 = data;
#endif
}	


void tlc5947_update(void)
{
	int8_t j, i;								// Index variables
	
	for (j=TLC5947_CHIPS-1; j>=0; j--)			// for any chip
	{
		for (i=35;i>=0;i--)						// transfer 36x8 = 288 = 12x24 bit
		{
			transmit_byte(grayscale[j][i]);
		}
	}
	
#if (TLC5947_USART==0)
	while(!(UCSR0A&(1<<TXC0)));					// wait for complete transfer
	UCSR0A |= (1<<TXC0);						// clear flag
#elif (TLC5947_USART==1)
	while(!(UCSR1A&(1<<TXC1)));		
	UCSR1A |= (1<<TXC1);			
#endif	
				
	SET_XLAT;									// pulse latch
	CLR_XLAT;
	
	SET_BLANK;									// pulse blank
	CLR_BLANK;									// to restart grayscale cycle
}


void tlc5947_set(uint8_t chip, uint8_t channel, uint16_t brightness)
{
	if (chip>=TLC5947_CHIPS) return;			// check for chip number
	if (channel>=24) return;					// check for channel number
	
	if (brightness>4095)						// limit brightness
	{
		brightness = 4095;
	}
	
	uint8_t index = (channel*3)/2;				// calculate index of 8-bit aligned array
	
	if ((channel%2) == 0)						// even channels aligning
	{
		grayscale[chip][index] = (uint8_t)brightness;
		grayscale[chip][index+1] &= 0b11110000;
		grayscale[chip][index+1] |= (uint8_t)(brightness >> 8);
	}
	else										// odd channels aligning
	{
		grayscale[chip][index] &= 0b00001111;
		grayscale[chip][index] |= (uint8_t)(brightness << 4);
		grayscale[chip][index+1] = (uint8_t)(brightness >> 4);
	}
}


void tlc5947_init(void)
{
	// initialize the pins
	TLC5947_PORT &= ~((1<<TLC5947_DATA)|(1<<TLC5947_SCLK)|(1<<TLC5947_XLAT)|(1<<TLC5947_BLANK));	
	TLC5947_DDR |= ((1<<TLC5947_DATA)|(1<<TLC5947_SCLK)|(1<<TLC5947_XLAT)|(1<<TLC5947_BLANK));
	
#if (TLC5947_USART==0)
	UBRR0  = 1;									// maximum speed
	UCSR0C = 0b11000000;						// MSPI mode, MSB first, PHASE=0, POL=0
	UCSR0B = 0b00001000;						// enable transmitter
#elif (TLC5947_USART==1)
	UBRR1  = 1;				
	UCSR1C = 0b11000000;
	UCSR1B = 0b00001000;
#endif
	
	tlc5947_update();							// update all LEDs (turn off)	
}


#if (TLC5947_CHIPS == 0)
	#error What do you want with this library without any chip connected? TLC5947_CHIPS have to be > 0.
#endif

#if (TLC5947_CHIPS > 1)
	#warning More than one chip is not tested yet!
#endif

#if (TLC5947_USART > 1)
	#error Only USART0 and USART1 are implemented!
#endif